/*********************************************************
 *
 * IMS_MIDI.c
 *
 * Authors: Ben McQuiston
 *
 * C function to convert our house IMS files into a usable MIDI file
 *
 *  Last Revision: October 25, 2006, 
 *			Major structure created for Mthd print and 
 *                         first MTrk with tempo specifiers 
 *                                          
 *  
 *****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

int
ImsToMIDI(char *filename)
{
	printf("Generating MIDI file...\n");
	
	int tempo;
	int divisor = 60000000;
	int mspqn;
	char zero = 0x00;

	FILE *fp1;
	FILE *fp2;

	char ch;
	char keySig;
	char timeSigTop;
	char timeSigBot;

	fp1 = fopen(filename, "rb");
	fp2 = fopen("output.mid","w");

	if(fp1 == NULL){
	  printf("ERROR\n");
	  return -1;
	}

	
/********************************************************
*
*  Mthd header to file...this will always be the same
*
*********************************************************/
		
	fprintf(fp2,"%c%c%c%c",0x4D,0x54,0x68,0x64); /* Mthd */
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",0x06);/* Always 6 bytes in Mthd header */
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",0x01);/* MIDI Type 1 */
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",0x02);/* Number of Tracks... could change?*/
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",0xF0); /*240 PPQN always..so dont forget*/

	 /*fprintf(fp2,"%c",zero); delta time 00 */

/********************************************************
*  
*   Mtrk Header...
*
*********************************************************/

	fprintf(fp2,"%c%c%c%c",0x4D,0x54,0x72,0x6B); /* Mtrk */
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",0x19); /* 25 bytes in first track...stays the same */

	fprintf(fp2,"%c",zero); /* delta time 00 */
	
/******************************************************
*
*  Start at beginning of IMS file to start getting 
*  musical information...
*
********************************************************/

	fseek(fp1,0,SEEK_SET);
	
	/* tempo cammand FF 51 03 xx xx xx = micros / quarter note */

	fscanf(fp1,"%c",&ch);
	tempo = (int) ch;
	mspqn = divisor / tempo;
	fprintf(fp2,"%c%c%c", 0xFF,0x51,0x03); /* tempo command */
	fprintf(fp2,"%c%c%c",*((char *) &mspqn +2),*((char *) &mspqn+1), *((char *) &mspqn));

	fprintf(fp2,"%c",zero);  /*delta time 00 */

	/* key sig command FF 59 02 xx xx = flats or sharps and major or minor */

	fscanf(fp1,"%c",&ch);
	keySig = ch;

	/* logic........for now we are just going to force it 
	* to C major no flats we are not sure why we need this...*/
	
	fprintf(fp2,"%c%c%c",0xFF,0x59,0x02);
	fprintf(fp2,"%c",zero); /* no flats (C) */
	fprintf(fp2,"%c",zero); /* major */ 
	fprintf(fp2,"%c",zero); /* delta time 00 */

	/* Time Signature command FF 58 04 xx xx xx xx 
	* numerator and denominator of time signature
	* number of MIDI clocks per beat (always 24)
	* number of 32nd notes per quater note (always 8)
	*/
	
	fscanf(fp1,"%c",&ch);
	timeSigTop = ch;
	fprintf(fp2,"%c%c%c",0xFF,0x58,0x04);

	fprintf(fp2,"%c",ch);

	fscanf(fp1,"%c",&ch);
	timeSigBot = ch;

	/* denominator must be a negative power of four,
	*  so we check if its a 4 or an 8 and print the corresponding
	* power to the file...
	*/

	if ((int) ch == 4){
		fprintf(fp2,"%c",0x02); 
	}
	else if ((int) ch == 8){
		fprintf(fp2,"%c",0x03);
	}

	fprintf(fp2,"%c%c",0x18,0x08); /* 24 MIDI clocks per beat,
								   * 8 32nds per quarter*/								   

	fprintf(fp2,"%c",zero);  /* delta time 00 */

	fprintf(fp2,"%c%c%c",0xFF,0x2F,0); /* End of Track Marker */


/*************************************************************
*
*  NOTE PROCESSING
*
*  Actually deal with the note values and lengths
*  Figure out how long it is going to be so that the Mtrk
*  header can tell how many bytes to expect, then just 
*  process all of the notes two bytes at a time...
*
************************************************************/
	
	fprintf(fp2,"%c%c%c%c",0x4D,0x54,0x72,0x6B); /* Mtrk */
	
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",0x17);

	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c%c",0xC0,0x00);
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c%c%c",0x90,0x3C,0x40);
	fprintf(fp2,"%c",0x3D);
	fprintf(fp2,"%c%c%c",0x80,0x3C,0x40);
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c%c%c",0x90,0x3E,0x40);
	fprintf(fp2,"%c",0x3D);
	fprintf(fp2,"%c%c%c",0x80,0x3E,0x40);
	fprintf(fp2,"%c",0x81);
	fprintf(fp2,"%c",0x28);
	fprintf(fp2,"%c%c%c",0xFF,0x2F,0); /* End of Track Marker */

	return 0;
	
}
